// Copyright 2017, the Flutter project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'package:meta/meta.dart';

import '../firebase_database.dart' show DataSnapshot, Event, Query;
import 'firebase_list.dart';

/// Sorts the results of `query` on the client side using to the `comparator`.
///
// TODO(jackson) We don't support children moving around.
// TODO(jackson) Right now this naively sorting the list after an insert.
// We can be smarter about how we handle insertion and take advantage of the
// fact that we know the list is already sorted. See example here:
// https://github.com/firebase/FirebaseUI-iOS/blob/master/FirebaseDatabaseUI/FUISortedArray.m
class ClientSortedList extends FirebaseList {
  ClientSortedList({
    @required Query query,
    @required this.comparator,
    this.onChildAdded,
    this.onChildRemoved,
    this.onChildChanged,
  }) : super(query: query) {
    assert(comparator != null);
    listen(query.onChildAdded, _onChildAdded);
    listen(query.onChildRemoved, _onChildRemoved);
    listen(query.onChildChanged, _onChildChanged);
  }

  /// The comparator used to sort the list on the client side
  final Comparator<DataSnapshot> comparator;

  /// Called when the child has been added
  final ChildCallback onChildAdded;

  /// Called when the child has been removed
  final ChildCallback onChildRemoved;

  /// Called when the child has changed
  final ChildCallback onChildChanged;

  void _onChildAdded(Event event) {
    snapshots.add(event.snapshot);
    snapshots.sort(comparator);
    onChildAdded(snapshots.indexOf(event.snapshot), event.snapshot);
  }

  void _onChildRemoved(Event event) {
    final DataSnapshot snapshot =
        snapshots.firstWhere((DataSnapshot snapshot) {
      return snapshot.key == event.snapshot.key;
    });
    final int index = snapshots.indexOf(snapshot);
    snapshots.removeAt(index);
    onChildRemoved(index, snapshot);
  }

  void _onChildChanged(Event event) {
    final DataSnapshot snapshot =
        snapshots.firstWhere((DataSnapshot snapshot) {
      return snapshot.key == event.snapshot.key;
    });
    final int index = snapshots.indexOf(snapshot);
    this[index] = event.snapshot;
    onChildChanged(index, event.snapshot);
  }
}
